//
// ViewerCanvas.java
//
// Here lies the code for the data display canvas...
//

package spg;

import java.awt.*;
import java.awt.event.*;

public class ViewerCanvas extends Canvas implements Runnable, MouseListener,
						MouseMotionListener {
    static int LEGEND_FONT_SIZE = 14;
    static int LABEL_FONT_SIZE = 12;

    static int HOVER_SECONDS = 2;	// Number of seconds before cursor hover activated

    ViewerState global;
    DataLayout currentLayout;

    private Thread cursorHoverThread;

    double dataPoint[];
    double tVal;

    Rectangle bounds;
    Dimension offDimension;
    Graphics offGraphics;
    Image offImage;
    Font legendFont, labelFont;
    FontMetrics legendFM, labelFM;

    int latestX[], latestY[];
    boolean currentlyHovering;

    public ViewerCanvas(ViewerState gInfo) {
	global = gInfo;
	addMouseListener(this);
	addMouseMotionListener(this);

	latestX = new int[HOVER_SECONDS];
	latestY = new int[HOVER_SECONDS];
	resetCursorHover(true);

	cursorHoverThread = new Thread(this, "cursorHover");
	cursorHoverThread.start();
    }

    public void run() {
	// This code will detect a cursor hover
	// (similar to tool-tips) and set up
	// cursorX and cursorY appropriately.
	// At all other times they should be
	// set to -1.
	Thread myThread = Thread.currentThread();
	boolean hoverDetected;
	while (cursorHoverThread == myThread) {
	    if ((latestX[0] != -1) && !currentlyHovering) {
		hoverDetected = true;
		for (int i = 1; i < HOVER_SECONDS; i++) {
		    if ((latestX[i] != latestX[0]) || (latestY[i] != latestY[0]))
			hoverDetected = false;
		}
	    
		if (hoverDetected) {
		    currentlyHovering = true;
		    global.current.cursorX = latestX[0];
		    global.current.cursorY = latestY[0];
		    global.canvas.repaint();
		} else {
		    pushLatestCursor(latestX[0], latestY[0]);
		}
	    }

	    try {
		Thread.sleep(1000);
	    } catch (InterruptedException e) {
	    }
	}
    }

    void resetCursorHover(boolean clearLatest) {
	if (global.current.cursorX != -1) {
	    global.current.cursorX = -1;
	    global.current.cursorY = -1;
	    global.canvas.repaint();
	}
	currentlyHovering = false;
	if (clearLatest) {
	    for (int i = 0; i < HOVER_SECONDS; i++)
		latestX[i] = latestY[i] = -1;
	}
    }
    
    void pushLatestCursor(int x, int y) {
	for (int i = HOVER_SECONDS - 1; i > 0; i--) {
	    latestX[i] = latestX[i - 1];
	    latestY[i] = latestY[i - 1];
	}
	latestX[0] = x;
	latestY[0] = y;
    }

    public synchronized void mouseClicked(MouseEvent ev) {
	resetCursorHover(true);
    }

    public synchronized void mouseEntered(MouseEvent ev) {
	resetCursorHover(true);
    }

    public synchronized void mouseExited(MouseEvent ev) {
	resetCursorHover(true);
    }

    public synchronized void mousePressed(MouseEvent ev) {
	resetCursorHover(true);

	global.current.brushOffset = -1.0;
	if (ev.getModifiers() != InputEvent.BUTTON1_MASK) {
	    global.current.brushOffsetX = -1;
	    global.canvas.repaint();
	} else {
	    if (this.contains(ev.getX(), ev.getY())) {
		global.current.brushOffsetX = ev.getX();
		global.canvas.repaint();
	    }
	}
    }

    public synchronized void mouseReleased(MouseEvent ev) {
	resetCursorHover(true);
    }

    public synchronized void mouseDragged(MouseEvent ev) {
	resetCursorHover(true);

	if (ev.getModifiers() == InputEvent.BUTTON1_MASK ||
	    ev.getModifiers() == 0)				// Work-around for win32!
	{
	    global.current.brushOffset = -1.0;
	    if (this.contains(ev.getX(), ev.getY())) {
		global.current.brushOffsetX = ev.getX();
		global.canvas.repaint();
	    }
	}
    }

    public synchronized void mouseMoved(MouseEvent ev) {
	// Here is where we'll keep track of
	// the cursor location for tool-tip type
	// purposes.  The primary use will be to
	// provide raw data to the user if she
	// hovers the cursor over a glyph for
	// some number of seconds...
	if (this.contains(ev.getX(), ev.getY())) {
	    resetCursorHover(false);
	    pushLatestCursor(ev.getX(), ev.getY());
	} else {
	    resetCursorHover(true);
	}
    }

    void noPlugins(Graphics g) {
	offGraphics.setFont(legendFont);
	offGraphics.setColor(Color.black);
	offGraphics.drawString("ERROR!  No plugins found!", 
	    bounds.x + (bounds.width - 
	    legendFM.stringWidth("ERROR!  No plugins found!")) / 2,
	    bounds.y + (bounds.height - (2 * legendFM.getHeight())));

	// Draw title graphics
	offGraphics.drawImage(global.titleImage, 
	    bounds.x + (bounds.width - global.titleImage.getWidth(this)) / 2, 
	    bounds.y + (bounds.height - global.titleImage.getHeight(this)) / 2,
	    this);

	// Swap buffers
	g.drawImage(offImage, 0, 0, this);
    }

    void noDataSet(Graphics g) {
	offGraphics.setFont(legendFont);
	offGraphics.setColor(Color.black);
	if (global.inAnApplet) {
	    offGraphics.drawString("No data set open!", 
		bounds.x + (bounds.width - 
		legendFM.stringWidth("No data set open!")) / 2,
		bounds.y + (bounds.height - (2 * legendFM.getHeight())));
	} else {
	    offGraphics.drawString("Please open a data set...", 
		bounds.x + (bounds.width - 
		legendFM.stringWidth("Please open a data set...")) / 2,
		bounds.y + (bounds.height - (2 * legendFM.getHeight())));
	}

	// Draw title graphics
	offGraphics.drawImage(global.titleImage, 
	    bounds.x + (bounds.width - global.titleImage.getWidth(this)) / 2, 
	    bounds.y + (bounds.height - global.titleImage.getHeight(this)) / 2,
	    this);

	// Swap buffers
	g.drawImage(offImage, 0, 0, this);
    }

    void noEnabledDim(Graphics g) {
	offGraphics.setFont(legendFont);
	offGraphics.setColor(Color.black);
	offGraphics.drawString("No dimensions enabled!", 
	    bounds.x + (bounds.width - 
	    legendFM.stringWidth("No dimensions enabled!")) / 2,
	    bounds.y + ((bounds.height - legendFM.getHeight()) / 2) +
	    legendFM.getAscent());

	// Swap buffers
	g.drawImage(offImage, 0, 0, this);
    }

    void getExtremeDataValues() {
	if (global.current.autoNormalize) {
	    currentLayout.greatestValue = 0.0;	// Pre-normed for each dim to [EPSILON, 1]
	    currentLayout.leastValue = 1.0;

	    for (int i = 0; i < global.constant.numDataPoints && 
		(currentLayout.greatestValue < 1.0 || currentLayout.leastValue > ViewerState.EPSILON); i++) 
	    {
		dataPoint = global.constant.normData[i];
		tVal = dataPoint[0];

		if ((tVal < global.current.tVal) ||
		    (tVal > global.derived.tVal2))
		    continue;

		for (int j = 0; j < global.current.enabledDim; j++) {
		    if (dataPoint[global.current.dimMap[j + 1]] > currentLayout.greatestValue)
			currentLayout.greatestValue = dataPoint[global.current.dimMap[j + 1]];
		    if (dataPoint[global.current.dimMap[j + 1]] < currentLayout.leastValue)
			currentLayout.leastValue = dataPoint[global.current.dimMap[j + 1]];
		}
	    }

	} else {
	    currentLayout.greatestValue = 1.0;
	    currentLayout.leastValue = ViewerState.EPSILON;
	}
    }

    void updateBackBuffer(Dimension d) {
        offDimension = d;
	offImage = createImage(d.width, d.height);
	offGraphics = offImage.getGraphics();

	legendFont = new Font("SansSerif", Font.BOLD, LEGEND_FONT_SIZE);
	labelFont = new Font("SansSerif", Font.BOLD, LABEL_FONT_SIZE);
	legendFM = offGraphics.getFontMetrics(legendFont);
	labelFM = offGraphics.getFontMetrics(labelFont);
	if (global.numLayoutPlugins > 0) {
	    currentLayout.setLabelFontInfo(labelFont, labelFM);
	}

        //
        // Garbage collection is not fast enough
        // to always keep up with us here, so
        // we'll force it to happen.
        //
        System.gc();
    }



    public void paint(Graphics g) {
	 update(g);
    }

    public synchronized void update(Graphics g) {
	Dimension d = getSize();

	// New layout may have new shape...
        if (global.layoutTypeDirty) {
	    currentLayout = global.layoutPlugins[global.current.layoutType];

	    updateBackground(g);
	    global.layoutTypeDirty = false;
	}

	// Initialize and respond to changes in the canvas geometry
	if ((offGraphics == null) || (d.width != offDimension.width) ||
	   (d.height != offDimension.height))
	{
	    updateBackBuffer(d);

	    // Get layout type dependent bounds
	    bounds = currentLayout.getLayoutBounds(d);
	}

	// Clear background
	offGraphics.setColor(Color.white);
	offGraphics.fillRect(bounds.x, bounds.y, bounds.width + 1, bounds.height + 1);

	offGraphics.setColor(Color.black);
	offGraphics.drawRect(bounds.x, bounds.y, bounds.width, bounds.height);

	// Set clipRect
	offGraphics.setClip(bounds.x + 1, bounds.y + 1, 
	    bounds.width - 1, bounds.height - 1);

	// If no plugins loaded, indicate this!
	if ((global.numLayoutPlugins == 0) || (global.numGlyphPlugins == 0)) {
	    noPlugins(g);
	    return;
	}

	// If no data set loaded, indicate this!
	if (global.fileName == null) {
	    noDataSet(g);
	    return;
	}

	// If no enabled dimensions, get the hell out of Dodge
	if (global.current.enabledDim == 0) {
	    noEnabledDim(g);
	    return;
	}

	// Prescan the data to determine the range
	getExtremeDataValues();

	// Layout dependent drawing
	currentLayout.drawLayout(offGraphics, global);

	// Swap buffers
	g.drawImage(offImage, 0, 0, this);
     }

    public synchronized void updateBackground(Graphics g) {
	Dimension d = getSize();

	// Clear background
	g.setColor(getBackground());
	g.fillRect(0, 0, d.width + 1, d.height + 1);

	offGraphics = null;
    }
}