//
// ViewerState.java
//
// This file contains all of the state structures and
// file I/O for loading and saving these structures.
//

package spg;

import java.awt.*;
import java.io.*;
import java.net.*;

public class ViewerState {
    static float VERSION = (float)1.3;		// File format version

    static double EPSILON = 0.01;		// Norm to [EPSILON,1]
    static double TSCROLLFACTOR = 100.0;

    boolean inAnApplet;

    int numLayoutPlugins;
    DataLayout layoutPlugins[];

    int numGlyphPlugins;
    DataGlyph glyphPlugins[];

    ViewerAdjustableState original;		// what is saved
    ViewerAdjustableState current;	// what is current
    ViewerStaticState constant;
    ViewerDerivedState derived;

    ViewerCanvas canvas;
    ViewerSideControls sideControls;
    ViewerBottomControls bottomControls;
    ViewerFrame frame;

    Image titleImage;

    boolean layoutTypeDirty;			// Always dirty to start

    String fileName;				// Implicit in the file

    boolean checkForSave() {
	// Compare current state with original state
	if (current.equalTo(original, constant.dataDim))
	    return true;

	// If things have changed, ask user to save
	Dimension d;
	Rectangle r;
	String message[] = new String[6];
	int just[] = new int[6];

	message[0] = "";
	message[1] = "  You have changed the viewing  ";
	message[2] = "  parameters for this data set.  ";
	message[3] = "";
	message[4] = "  Would you like to save them?  ";
	message[5] = "";
	just[0] = just[1] = just[2] = just[3] = just[4] = just[5] = Label.CENTER;

	YesNoCancelDialog queryDialog = new YesNoCancelDialog(frame, 
	    "Save data set?", 6, message, just, true);
	queryDialog.setResizable(false);
	r = frame.getBounds();
	d = queryDialog.getSize();
	queryDialog.setLocation((r.width - d.width) / 2 + r.x, 
	    (r.height - d.height) / 2 + r.y);
	queryDialog.pack();
	queryDialog.show();

	frame.toFront();

	// Return false if user cancels action
	if (queryDialog.result == YesNoCancelDialog.CANCEL)
	    return false;
	else {
	    if (queryDialog.result == YesNoCancelDialog.YES) {
		// Save using current fileName, no dialogs
		saveState();
	    }

	    return true;
	}
    }

    void loadState(String newFileName) {
	boolean successfulLoad = false;
	String errorString;
	float thisVersion;
	ViewerState newGlobal = new ViewerState();

	// Grab state from text file using current fileName
	try {	do {
	    Reader r;

	    if (inAnApplet) {
		URL dataSet = new URL(newFileName);
		r = new BufferedReader(new InputStreamReader(dataSet.openStream()));
	    } else {
		File inputFile = new File(newFileName);
		r = new FileReader(inputFile);
	    }

	    StreamTokenizer st = new StreamTokenizer(r);

	    st.slashStarComments(true);
	    st.slashSlashComments(true);
	    st.commentChar('#');
	    st.whitespaceChars(',', ',');
	    st.whitespaceChars(':', ':');

	    errorString = "Error when parsing: Version";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("Version"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    // Currently there are four versions (1.[0123]) in existence
	    thisVersion = (float)st.nval;
	    if (thisVersion != (float)1.0 && 
		thisVersion != (float)1.1 &&
		thisVersion != (float)1.2 &&
		thisVersion != (float)1.3)
		break;

	    errorString = "Error when parsing: tMinimumUnit";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("tMinimumUnit"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    newGlobal.constant.tMinimumUnit = st.nval;

	    errorString = "Error when parsing: tMinValue";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("tMinValue"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    newGlobal.constant.tMin = st.nval;
	    if (newGlobal.constant.tMin < newGlobal.constant.tMinimumUnit)
		break;

	    errorString = "Error when parsing: tMaxValue";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("tMaxValue"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    newGlobal.constant.tMax = st.nval;
	    if (newGlobal.constant.tMax < newGlobal.constant.tMin)
		break;

	    errorString = "Error when parsing: tValue";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("tValue"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    newGlobal.original.tVal = st.nval;
	    if ((newGlobal.original.tVal < newGlobal.constant.tMin) || 
		(newGlobal.original.tVal > newGlobal.constant.tMax))
		break;

	    errorString = "Error when parsing: CycleLength";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("CycleLength"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    newGlobal.original.cycleLength = st.nval;
	    newGlobal.constant.tSpan = newGlobal.constant.tMax - 
		newGlobal.constant.tMin + newGlobal.constant.tMinimumUnit;
	    if ((newGlobal.original.cycleLength < newGlobal.constant.tMinimumUnit) || 
		(newGlobal.original.cycleLength > newGlobal.constant.tSpan))
		break;

	    errorString = "Error when parsing: VisibleCycles";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("VisibleCycles"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    newGlobal.original.visibleCycles = (int)st.nval;
	    newGlobal.derived.numCycles = (int)Math.ceil(newGlobal.constant.tSpan / 
		newGlobal.original.cycleLength);
	    if ((newGlobal.original.visibleCycles < 1) || 
		(newGlobal.original.visibleCycles > newGlobal.derived.numCycles))
		break;

	    errorString = "Error when parsing: GlyphScale";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("GlyphScale"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    newGlobal.original.glyphScale = st.nval;
	    if ((newGlobal.original.glyphScale < 0.0) || 
		(newGlobal.original.glyphScale > 2.0))
		break;

	    if (thisVersion <= (float)1.2) {
		errorString = "Error when parsing: GlyphType";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("GlyphType"))
		    break;
		if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		    break;

	    	// Translate old built-in glyph
		// type to new plugin if it
		// exists.
		int oldGlyphType = (int)st.nval;
		String pluginName;
		if (oldGlyphType == 1)
		    pluginName = "Star Glyph";
		else if (oldGlyphType == 2)
		    pluginName = "Bar Glyph";
		else if (oldGlyphType == 4)
		    pluginName = "Pie Glyph";
		else break;
		oldGlyphType = -1;
		for (int i = 0; i < numGlyphPlugins; i++) {
		    if (glyphPlugins[i].getName() == pluginName) {
			oldGlyphType = i;
			break;
		    }
		}
		if (oldGlyphType != -1)
		    newGlobal.original.glyphType = oldGlyphType;
		else
		    // Plugin not available... weird...
		    newGlobal.original.glyphType = 0;
	    }

	    if (thisVersion <= (float)1.2) {
		errorString = "Error when parsing: ViewType";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("ViewType"))
		    break;
		if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		    break;

	    	// Translate old built-in layout
		// type to new plugin if it
		// exists.
		int oldLayoutType = (int)st.nval;
		String pluginName;
		if (oldLayoutType == 1)
		    pluginName = "Spiral Layout";
		else if (oldLayoutType == 2)
		    pluginName = "Linear Layout";
		else if (oldLayoutType == 4)
		    pluginName = "Stacked Layout";
		else break;
		oldLayoutType = -1;
		for (int i = 0; i < numLayoutPlugins; i++) {
		    if (layoutPlugins[i].getName() == pluginName) {
			oldLayoutType = i;
			break;
		    }
		}
		if (oldLayoutType != -1)
		    newGlobal.original.layoutType = oldLayoutType;
		else
		    // Plugin not available... weird...
		    newGlobal.original.layoutType = 0;
	    }

	    if (thisVersion >= (float)1.2) {
		errorString = "Error when parsing: BrushOffset";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("BrushOffset"))
		    break;
		if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		    break;
		newGlobal.original.brushOffset = st.nval;
		if (newGlobal.original.brushOffset != -1)
		    newGlobal.original.brushOffsetX = 0;
		else
		    newGlobal.original.brushOffsetX = -1;
	    } else {
		newGlobal.original.brushOffset = -1.0;
		newGlobal.original.brushOffsetX = -1;
	    }

	    errorString = "Error when parsing: AutoNormalize";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("AutoNormalize"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    if (st.nval == 1.0)
		newGlobal.original.autoNormalize = true;
	    else if (st.nval == 0.0)
		newGlobal.original.autoNormalize = false;
	    else
		break;

	    errorString = "Error when parsing: ShowLabels";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("ShowLabels"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    if (st.nval == 1.0)
		newGlobal.original.showLabels = true;
	    else if (st.nval == 0.0)
		newGlobal.original.showLabels = false;
	    else
		break;

	    errorString = "Error when parsing: ShowGuides";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("ShowGuides"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    if (st.nval == 1.0)
		newGlobal.original.showGuidelines = true;
	    else if (st.nval == 0.0)
		newGlobal.original.showGuidelines = false;
	    else
		break;

	    if (thisVersion >= (float)1.2) {
		errorString = "Error when parsing: FillGlyphs";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("FillGlyphs"))
		    break;
		if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		    break;
		if (st.nval == 1.0)
		    newGlobal.original.fillGlyphs = true;
		else if (st.nval == 0.0)
		    newGlobal.original.fillGlyphs = false;
		else
		    break;
	    } else {
		newGlobal.original.fillGlyphs = true;
	    }

	    if (thisVersion == (float)1.0) {
		errorString = "Error when parsing: ConstantColor";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("ConstantColor"))
		    break;
		if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		    break;
		if (st.nval == 1.0)
		    newGlobal.original.colorPerGlyph = false;
		else if (st.nval == 0.0)
		    newGlobal.original.colorPerGlyph = true;
		else
		    break;
	    } else {
		errorString = "Error when parsing: ColorPerGlyph";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("ColorPerGlyph"))
		    break;
		if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		    break;
		if (st.nval == 1.0)
		    newGlobal.original.colorPerGlyph = true;
		else if (st.nval == 0.0)
		    newGlobal.original.colorPerGlyph = false;
		else
		    break;
	    }

	    if (thisVersion <= (float)1.2) {
		errorString = "Error when parsing: ShowSpiral";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("ShowSpiral"))
		    break;
		if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		    break;

		// If Spiral Layout plugin is
		// installed, we'll put this
		// defunct setting to use.
		boolean oldShowSpiral;
		if (st.nval == 1.0)
		    oldShowSpiral = true;
		else if (st.nval == 0.0)
		    oldShowSpiral = false;
		else
		    break;
		for (int i = 0; i < numLayoutPlugins; i++) {
		    if (layoutPlugins[i].getName() == "Spiral Layout") {
			int numOptions = layoutPlugins[i].getNumOptions();
			for (int j = 0; j < numOptions; j++) {
			    String optionName = layoutPlugins[i].getOptionName(j);
			    if (optionName == "Show Spiral") {
				layoutPlugins[i].setOptionValue(j, oldShowSpiral);
				break;
			    }
			}
			break;
		    }
		}
	    }

	    if (thisVersion <= (float)1.2) {
		errorString = "Error when parsing: InToOut";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("InToOut"))
		    break;
		if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		    break;

		// If Spiral Layout plugin is
		// installed, we'll put this
		// defunct setting to use.
		boolean oldInToOutStatus;
		if (st.nval == 1.0)
		    oldInToOutStatus = true;
		else if (st.nval == 0.0)
		    oldInToOutStatus = false;
		else
		    break;
		for (int i = 0; i < numLayoutPlugins; i++) {
		    if (layoutPlugins[i].getName() == "Spiral Layout") {
			int numOptions = layoutPlugins[i].getNumOptions();
			for (int j = 0; j < numOptions; j++) {
			    String optionName = layoutPlugins[i].getOptionName(j);
			    if (optionName == "Spiral Outward") {
				layoutPlugins[i].setOptionValue(j, oldInToOutStatus);
				break;
			    }
			}
			break;
		    }
		}
	    }

	    if (thisVersion <= (float)1.2) {
		errorString = "Error when parsing: ShowLinearPlot";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("ShowLinearPlot"))
		    break;
		if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		    break;

		// If Linear Layout plugin is
		// installed, we'll put this
		// defunct setting to use.
		boolean oldShowLinearPlot;
		if (st.nval == 1.0)
		    oldShowLinearPlot = true;
		else if (st.nval == 0.0)
		    oldShowLinearPlot = false;
		else
		    break;
		for (int i = 0; i < numLayoutPlugins; i++) {
		    if (layoutPlugins[i].getName() == "Linear Layout") {
			int numOptions = layoutPlugins[i].getNumOptions();
			for (int j = 0; j < numOptions; j++) {
			    String optionName = layoutPlugins[i].getOptionName(j);
			    if (optionName == "Show Plot") {
				layoutPlugins[i].setOptionValue(j, oldShowLinearPlot);
				break;
			    }
			}
			break;
		    }
		}
	    }

	    errorString = "Error when parsing: TotalDims";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("TotalDims"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    newGlobal.constant.dataDim = (int)st.nval;
	    if (newGlobal.constant.dataDim < 1)
		break;

	    errorString = "Error when parsing: EnabledDims";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("EnabledDims"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    newGlobal.original.enabledDim = (int)st.nval;
	    if ((newGlobal.original.enabledDim < 1) || 
		(newGlobal.original.enabledDim > newGlobal.constant.dataDim))
		break;

	    errorString = "Error when parsing: NumDataPoints";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("NumDataPoints"))
		break;
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		break;
	    newGlobal.constant.numDataPoints = (int)st.nval;
	    if (newGlobal.constant.numDataPoints < 1)
		break;

	    if (thisVersion >= (float)1.2) {
		errorString = "Error when parsing: DimMap";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("DimMap"))
		    break;
		newGlobal.original.dimMap = new int[newGlobal.constant.dataDim + 1];
		successfulLoad = true;
		for (int i = 0; i <= newGlobal.original.enabledDim; i++) {
		    if (st.nextToken() != StreamTokenizer.TT_NUMBER) {
			successfulLoad = false;
			break;
		    }
		    newGlobal.original.dimMap[i] = (int)st.nval;
		    if ((newGlobal.original.dimMap[i] < 0) || 
			(newGlobal.original.dimMap[i] > newGlobal.constant.dataDim)) {
			successfulLoad = false;
			break;
		    }
		}
		if (!successfulLoad)
		    break;
		successfulLoad = false;
	    } else {
		errorString = "Error when parsing: DimMap";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("DimMap"))
		    break;
		newGlobal.original.dimMap = new int[newGlobal.constant.dataDim + 1]; 
		successfulLoad = true;
		for (int i = 0; i <= newGlobal.constant.dataDim; i++) {
		    if (st.nextToken() != StreamTokenizer.TT_NUMBER) {
			successfulLoad = false;
			break;
		    }
		    newGlobal.original.dimMap[i] = (int)st.nval;
		    if ((newGlobal.original.dimMap[i] < 0) || 
			(newGlobal.original.dimMap[i] > newGlobal.constant.dataDim)) {
			successfulLoad = false;
			break;
		    }
		}
		if (!successfulLoad)
		    break;
		successfulLoad = false;
	    }

	    if (thisVersion >= (float)1.2) {
		errorString = "Error when parsing: ListMap";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("ListMap"))
		    break;
		newGlobal.original.listMap = new int[newGlobal.constant.dataDim]; 
		successfulLoad = true;
		for (int i = 0; i < newGlobal.constant.dataDim; i++) {
		    if (st.nextToken() != StreamTokenizer.TT_NUMBER) {
			successfulLoad = false;
			break;
		    }
		    newGlobal.original.listMap[i] = (int)st.nval;
		    if ((newGlobal.original.listMap[i] < 0) || 
			(newGlobal.original.listMap[i] > newGlobal.constant.dataDim - 1)) {
			successfulLoad = false;
			break;
		    }
		}
		if (!successfulLoad)
		    break;
		successfulLoad = false;
	    } else {
		newGlobal.original.listMap = new int[newGlobal.constant.dataDim]; 
		for (int i = 0; i < newGlobal.constant.dataDim; i++)
		    newGlobal.original.listMap[i] = i;
	    }

	    errorString = "Error when parsing: DimLabels";
	    if (st.nextToken() != StreamTokenizer.TT_WORD)
		break;
	    if (!st.sval.equalsIgnoreCase("DimLabels"))
		break;
	    newGlobal.constant.dimLabel = new String[newGlobal.constant.dataDim];
	    successfulLoad = true;
	    int newToken;
	    for (int i = 0; i < newGlobal.constant.dataDim; i++) {
		if ((newToken = st.nextToken()) != StreamTokenizer.TT_WORD &&
		    newToken != '"') 
		{
		    successfulLoad = false;
		    break;
		}
		newGlobal.constant.dimLabel[i] = st.sval;
	    }
	    if (!successfulLoad)
		break;
	    successfulLoad = false;

	    if (thisVersion >= (float)1.3) {
		errorString = "Error when parsing: CurrentLayout";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("CurrentLayout"))
		    break;
		if ((newToken = st.nextToken()) != StreamTokenizer.TT_WORD &&
		    newToken != '"')
		    break;
		String layoutName = st.sval;

		// If specified layout plugin does not exist,
		// use the safe value of zero.
		newGlobal.original.layoutType = 0;
		for (int i = 0; i < numLayoutPlugins; i++) {
		    if (layoutPlugins[i].getName().equalsIgnoreCase(layoutName)) {
			newGlobal.original.layoutType = i;
			break;
		    }
		}
	    }

	    if (thisVersion >= (float)1.3) {
		errorString = "Error when parsing: LayoutPluginsData";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("LayoutPluginsData"))
		    break;
		if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		    break;
		int numPlugins = (int)st.nval;
		successfulLoad = true;
		for (int i = 0; i < numPlugins; i++) {
		    if ((newToken = st.nextToken()) != StreamTokenizer.TT_WORD &&
			newToken != '"')
		    {
			successfulLoad = false;
			break;
		    }
		    String pluginName = st.sval;
		    int pluginIndex = -1;
		    for (int j = 0; j < numLayoutPlugins; j++) {
			if (pluginName.equalsIgnoreCase(layoutPlugins[j].getName())) {
			    pluginIndex = j;
			    break;
			}
		    }
		    successfulLoad = false;
		    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
			break;
		    successfulLoad = true;
		    int numOptions = (int)st.nval;
		    for (int j = 0; j < numOptions; j++) {
			if ((newToken = st.nextToken()) != StreamTokenizer.TT_WORD &&
			    newToken != '"')
			{
			    successfulLoad = false;
			    break;
			}
			String optionName = st.sval;
			successfulLoad = false;
			if (st.nextToken() != StreamTokenizer.TT_NUMBER)
			    break;
			successfulLoad = true;
			if (pluginIndex != -1) {
			    boolean optionValue = ((int)st.nval == 0) ? false : true;
			    // Set this option if it exists
			    for (int k = 0; k < layoutPlugins[pluginIndex].getNumOptions(); k++) {
				if (layoutPlugins[pluginIndex].getOptionName(k).equalsIgnoreCase(optionName)) {
				    layoutPlugins[pluginIndex].setOptionValue(k, optionValue);
				    break;
				}
			    }
			}
		    }
		    if (!successfulLoad)
			break;
		}
		if (!successfulLoad)
		    break;
		successfulLoad = false;
	    }

	    if (thisVersion >= (float)1.3) {
		errorString = "Error when parsing: CurrentGlyph";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("CurrentGlyph"))
		    break;
		if ((newToken = st.nextToken()) != StreamTokenizer.TT_WORD &&
		    newToken != '"')
		    break;
		String glyphName = st.sval;

		// If specified glyph plugin does not exist,
		// use the safe value of zero.
		newGlobal.original.glyphType = 0;
		for (int i = 0; i < numGlyphPlugins; i++) {
		    if (glyphPlugins[i].getName().equalsIgnoreCase(glyphName)) {
			newGlobal.original.glyphType = i;
			break;
		    }
		}
	    }

	    if (thisVersion >= (float)1.3) {
		errorString = "Error when parsing: GlyphPluginsData";
		if (st.nextToken() != StreamTokenizer.TT_WORD)
		    break;
		if (!st.sval.equalsIgnoreCase("GlyphPluginsData"))
		    break;
		if (st.nextToken() != StreamTokenizer.TT_NUMBER)
		    break;
		int numPlugins = (int)st.nval;
		successfulLoad = true;
		for (int i = 0; i < numPlugins; i++) {
		    if ((newToken = st.nextToken()) != StreamTokenizer.TT_WORD &&
			newToken != '"')
		    {
			successfulLoad = false;
			break;
		    }
		    String pluginName = st.sval;
		    int pluginIndex = -1;
		    for (int j = 0; j < numGlyphPlugins; j++) {
			if (pluginName.equalsIgnoreCase(glyphPlugins[j].getName())) {
			    pluginIndex = j;
			    break;
			}
		    }
		    successfulLoad = false;
		    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
			break;
		    successfulLoad = true;
		    int numOptions = (int)st.nval;
		    for (int j = 0; j < numOptions; j++) {
			if ((newToken = st.nextToken()) != StreamTokenizer.TT_WORD &&
			    newToken != '"')
			{
			    successfulLoad = false;
			    break;
			}
			String optionName = st.sval;
			successfulLoad = false;
			if (st.nextToken() != StreamTokenizer.TT_NUMBER)
			    break;
			successfulLoad = true;
			if (pluginIndex != -1) {
			    boolean optionValue = ((int)st.nval == 0) ? false : true;
			    // Set this option if it exists
			    for (int k = 0; k < glyphPlugins[pluginIndex].getNumOptions(); k++) {
				if (glyphPlugins[pluginIndex].getOptionName(k).equalsIgnoreCase(optionName)) {
				    glyphPlugins[pluginIndex].setOptionValue(k, optionValue);
				    break;
				}
			    }
			}
		    }
		    if (!successfulLoad)
			break;
		}
		if (!successfulLoad)
		    break;
		successfulLoad = false;
	    }

	    errorString = "Error: Data block not found!  ";
	    if ((newToken = st.nextToken()) != StreamTokenizer.TT_WORD) {
		errorString += "Instead we got: " + newToken + " for a token type!";
		break;
	    }
	    if (!st.sval.equalsIgnoreCase("Data")) {
		errorString += "Word token is \"" + st.sval + "\", not \"Data\"!";
		break;
	    }
	    newGlobal.constant.data = new double[newGlobal.constant.numDataPoints][newGlobal.constant.dataDim + 1];
	    newGlobal.constant.normData = new double[newGlobal.constant.numDataPoints][newGlobal.constant.dataDim + 1];
	    successfulLoad = true;
	    errorString = "Error when parsing: Data ";
	    for (int i = 0; i < newGlobal.constant.numDataPoints; i++) {
		for (int j = 0; j <= newGlobal.constant.dataDim; j++) {
		    st.nextToken();
		    if (st.ttype != StreamTokenizer.TT_NUMBER) {
			errorString += "(data line " + (i + 1) + ", field " 
			    + (j + 1) + ")";
			successfulLoad = false;
			break;
		    }
		    newGlobal.constant.data[i][j] = st.nval;
		}
		if (!successfulLoad)
		    break;
	    }
	    if (!successfulLoad)
		break;

	    r.close();

	} while(false);
	} catch (FileNotFoundException e) {
	    errorString = "Specified file not found!";
	} catch (Exception e) {
	    System.err.println("Couldn't load data:" + e);
	    errorString = "Error opening file!";
	}

	if (!successfulLoad) {
	    // Error dialog goes here
	    Dimension d;
	    Rectangle r;
	    String message[] = new String[7];
	    int just[] = new int[7];

	    message[0] = "";
	    message[1] = "  The specified data set does  ";
	    message[2] = "  not exist, or is of an  ";
	    message[3] = "  unrecognized format:  ";
	    message[4] = "";
	    message[5] = errorString;
	    message[6] = "";
	    just[0] = just[1] = just[2] = just[3] = just[4] = 
		just[5] = just[6] = Label.CENTER;

	    OkayDialog errorDialog = new OkayDialog(frame, 
		"Unable to Open File!", 7, message, just, true);
	    errorDialog.setResizable(false);
	    r = frame.getBounds();
	    d = errorDialog.getSize();
	    errorDialog.setLocation((r.width - d.width) / 2 + r.x, 
		(r.height - d.height) / 2 + r.y);
	    errorDialog.pack();
	    errorDialog.show();

	    frame.toFront();

	    return;
	}

	// Calculate normalized data (each dim to [EPSILON, 1])
	double leastSoFar, greatestSoFar, invScale;
	for (int j = 1; j <= newGlobal.constant.dataDim; j++) {
	    greatestSoFar = Double.NEGATIVE_INFINITY;
	    leastSoFar = Double.POSITIVE_INFINITY;
	    for (int i = 0; i < newGlobal.constant.numDataPoints; i++) {
		if (newGlobal.constant.data[i][j] > greatestSoFar)
		    greatestSoFar = newGlobal.constant.data[i][j];
		if (newGlobal.constant.data[i][j] < leastSoFar)
		    leastSoFar = newGlobal.constant.data[i][j];
	    }

	    invScale = (1.0 - EPSILON) / (greatestSoFar - leastSoFar);
	    for (int i = 0; i < newGlobal.constant.numDataPoints; i++) {
		newGlobal.constant.normData[i][j] = 
		    ((newGlobal.constant.data[i][j] - leastSoFar) * invScale) + EPSILON;
	    }
	}
	for (int i = 0; i < newGlobal.constant.numDataPoints; i++) {
	    newGlobal.constant.normData[i][0] = newGlobal.constant.data[i][0];
	}

	// Switch over to the new state!
	constant = newGlobal.constant;
	derived = newGlobal.derived;
	original = newGlobal.original;
	current.copyFrom(original, constant.dataDim);

	// Set UI component values as necessary
	fileName = newFileName;
	frame.setTitle("SpiralGlyphics - " + fileName);

	if (!inAnApplet) {
	    frame.saveMenuItem.setEnabled(true);
	    frame.saveAsMenuItem.setEnabled(true);
	}
	frame.glyphMenu.setEnabled(true);
	frame.layoutMenu.setEnabled(true);
	frame.optionsMenu.setEnabled(true);

	frame.labelsCheckboxMenuItem.setState(current.showLabels);
	frame.autoNormCheckboxMenuItem.setState(current.autoNormalize);
	frame.guidelinesCheckboxMenuItem.setState(current.showGuidelines);
	frame.fillGlyphsCheckboxMenuItem.setState(current.fillGlyphs);
	frame.colorPerDimCheckboxMenuItem.setState(!current.colorPerGlyph);
	frame.colorPerGlyphCheckboxMenuItem.setState(current.colorPerGlyph);

	bottomControls.tScrollbar.setEnabled(true);
	for (int i = 0; i < numLayoutPlugins; i++) {
	    frame.layoutCheckboxMenuItems[i].setState(current.layoutType == i);

	    // Set state and enable layout-specific options
	    int numOptions = layoutPlugins[i].getNumOptions();
	    for (int j = 0; j < numOptions; j++) {
		frame.layoutSpecificCheckboxMenuItems[i][j].setState(
		    layoutPlugins[i].getOptionValue(j));
		frame.layoutSpecificCheckboxMenuItems[i][j].setEnabled(current.layoutType == i);
	    }
	    sideControls.layoutTypeGroup.setSelectedCheckbox(
		sideControls.layoutTypeBoxes[current.layoutType]);
	    sideControls.layoutTypeBoxes[i].setEnabled(true);
	}
	for (int i = 0; i < numGlyphPlugins; i++) {
	    frame.glyphCheckboxMenuItems[i].setState(current.glyphType == i);
	    // Set state and enable glyph-specific options
	    int numOptions = glyphPlugins[i].getNumOptions();
	    for (int j = 0; j < numOptions; j++) {
		frame.glyphSpecificCheckboxMenuItems[i][j].setState(
		    glyphPlugins[i].getOptionValue(j));
		frame.glyphSpecificCheckboxMenuItems[i][j].setEnabled(current.glyphType == i);
	    }
	    sideControls.glyphTypeGroup.setSelectedCheckbox(
		sideControls.glyphTypeBoxes[current.glyphType]);
	    sideControls.glyphTypeBoxes[i].setEnabled(true);
	}
	sideControls.visibleCyclesScrollbar.setEnabled(true);
	sideControls.cycleLengthScrollbar.setEnabled(true);
	sideControls.glyphScaleScrollbar.setEnabled(true);
	sideControls.vcsVal.setEnabled(true);
	sideControls.clsVal.setEnabled(true);
	sideControls.gssVal.setEnabled(true);

	sideControls.dimLabelsPanel.removeAll();
	sideControls.dimensionLabelBox = new Checkbox[constant.dataDim];
	sideControls.dimensionUpButton = new Button[constant.dataDim];
	sideControls.dimensionDownButton = new Button[constant.dataDim];
	sideControls.dimLabelsC.weighty = 0.0;
	sideControls.dimLabelsC.gridheight = 1;
	sideControls.dimLabelsC.fill = GridBagConstraints.NONE;
	sideControls.dimLabelsC.anchor = GridBagConstraints.NORTHWEST;
	boolean thisDimEnabled;
	for (int i = 0; i < constant.dataDim; i++) {
	    sideControls.dimensionUpButton[i] = new Button("+");
	    sideControls.dimensionDownButton[i] = new Button("-");

	    sideControls.dimLabelsC.gridwidth = 1;
	    sideControls.dimLabelsGridBag.setConstraints(
		sideControls.dimensionUpButton[i], 
		sideControls.dimLabelsC);
	    sideControls.dimLabelsGridBag.setConstraints(
		sideControls.dimensionDownButton[i], 
		sideControls.dimLabelsC);
	    sideControls.dimensionUpButton[i].addActionListener(sideControls);
	    sideControls.dimensionDownButton[i].addActionListener(sideControls);
	    sideControls.dimLabelsPanel.add(
		sideControls.dimensionUpButton[i]);
	    sideControls.dimLabelsPanel.add(
		sideControls.dimensionDownButton[i]);

	    thisDimEnabled = false;
	    for (int j = 0; j < current.enabledDim; j++)
		if (current.dimMap[j + 1] == current.listMap[i] + 1)
		    thisDimEnabled = true;

	    if (thisDimEnabled) {
		sideControls.dimensionLabelBox[i] = new Checkbox("" + (i + 1) + ". " +
		    constant.dimLabel[current.listMap[i]], true);
	    } else {
		sideControls.dimensionLabelBox[i] = new Checkbox("    " +
		    constant.dimLabel[current.listMap[i]], false);
		sideControls.dimensionUpButton[i].setEnabled(false);
		sideControls.dimensionDownButton[i].setEnabled(false);
	    }

	    sideControls.dimLabelsC.gridwidth = GridBagConstraints.REMAINDER;
	    sideControls.dimLabelsGridBag.setConstraints(
		sideControls.dimensionLabelBox[i], 
		sideControls.dimLabelsC);
	    sideControls.dimensionLabelBox[i].addItemListener(sideControls);
	    sideControls.dimLabelsPanel.add(
		sideControls.dimensionLabelBox[i]);
	}
	sideControls.dimensionDownButton[0].setEnabled(false);
	sideControls.dimensionUpButton[constant.dataDim - 1].setEnabled(false);
	Label dummy = new Label("");
	sideControls.dimLabelsC.weighty = 1.0;
	sideControls.dimLabelsC.gridheight = GridBagConstraints.REMAINDER;
	sideControls.dimLabelsC.fill = GridBagConstraints.BOTH;
	sideControls.dimLabelsGridBag.setConstraints(dummy, 
	    sideControls.dimLabelsC);
	sideControls.dimLabelsPanel.add(dummy);
	sideControls.glyphLegendPanel.validate();

	int vcsShow = (int)Math.ceil((double)derived.numCycles / 5.0);
	sideControls.visibleCyclesScrollbar.setValues(current.visibleCycles, vcsShow, 1,
	    derived.numCycles + vcsShow);
	sideControls.visibleCyclesScrollbar.setBlockIncrement(vcsShow); 

	int clsMaxVal = (int)Math.rint(constant.tSpan * TSCROLLFACTOR);
	int clsShow = (int)Math.ceil((double)clsMaxVal / 5.0);
	sideControls.cycleLengthScrollbar.setValues(
	  (int)Math.rint(current.cycleLength * TSCROLLFACTOR), 
	  clsShow, 
	  (int)Math.rint(constant.tMinimumUnit * TSCROLLFACTOR), 
	  clsMaxVal + clsShow); 
	sideControls.cycleLengthScrollbar.setBlockIncrement((int)Math.rint(TSCROLLFACTOR));

	sideControls.glyphScaleScrollbar.setValues(
	  (int)Math.rint(current.glyphScale * 100.0), 40, 0, 240);

	sideControls.vcsVal.setText("" + current.visibleCycles);
	sideControls.vcsMax.setText("" + derived.numCycles);
	sideControls.clsMin.setText("" + constant.tMinimumUnit);
	sideControls.clsVal.setText("" + current.cycleLength);
	sideControls.clsMax.setText("" + constant.tSpan);
	sideControls.gssVal.setText("" + current.glyphScale);

	derived.tInterval = current.cycleLength * (double)current.visibleCycles;
	if (derived.tInterval > constant.tSpan)
	    derived.tInterval = constant.tSpan;
	double tTempMax = constant.tMax - derived.tInterval + constant.tMinimumUnit;
	int tMaxVal = (int)Math.rint(tTempMax * TSCROLLFACTOR);
	if (current.tVal > tTempMax)
	    current.tVal = ViewerState.massageDouble(tTempMax, TSCROLLFACTOR);
	derived.tVal2 = ViewerState.massageDouble(current.tVal +
	    derived.tInterval, TSCROLLFACTOR);
	bottomControls.tValLabel.setText("[" + current.tVal + ", " + derived.tVal2 + "]"); 
	bottomControls.tMin.setText("" + constant.tMin);
	bottomControls.tMax.setText("" + constant.tMax);

	int tShow = (int)Math.rint(derived.tInterval * TSCROLLFACTOR);
	bottomControls.tScrollbar.setValues(
	    (int)Math.rint(current.tVal * TSCROLLFACTOR), 
	    tShow, 
	    (int)Math.rint(constant.tMin * TSCROLLFACTOR), 
	    tMaxVal + tShow); 
	bottomControls.tScrollbar.setBlockIncrement(
	    (int)Math.rint(current.cycleLength * TSCROLLFACTOR));

	layoutTypeDirty = true;
    }

    void saveState() {
	// Dump state into text file using current fileName
	try {
	    File outputFile = new File(fileName);
	    FileWriter fw = new FileWriter(outputFile);

	    fw.write("Version: " + VERSION + '\r' + '\n');
	    fw.write("tMinimumUnit: " + constant.tMinimumUnit + '\r' + '\n');
	    fw.write("tMinValue: " + constant.tMin + '\r' + '\n');
	    fw.write("tMaxValue: " + constant.tMax + '\r' + '\n');
	    fw.write("tValue: " + current.tVal + '\r' + '\n');
	    fw.write("CycleLength: " + current.cycleLength + '\r' + '\n');
	    fw.write("VisibleCycles: " + current.visibleCycles + '\r' + '\n');
	    fw.write("GlyphScale: " + current.glyphScale + '\r' + '\n');
	    fw.write("BrushOffset: " + ((current.brushOffsetX == -1) ? -1.0 : current.brushOffset) + '\r' + '\n');
	    fw.write("AutoNormalize: " + (current.autoNormalize ? 1 : 0) + '\r' + '\n');
	    fw.write("ShowLabels: " + (current.showLabels ? 1 : 0) + '\r' + '\n');
	    fw.write("ShowGuides: " + (current.showGuidelines ? 1 : 0) + '\r' + '\n');
	    fw.write("FillGlyphs: " + (current.fillGlyphs ? 1 : 0) + '\r' + '\n');
	    fw.write("ColorPerGlyph: " + (current.colorPerGlyph ? 1 : 0) + '\r' + '\n');
	    fw.write("TotalDims: " + constant.dataDim + '\r' + '\n');
	    fw.write("EnabledDims: " + current.enabledDim + '\r' + '\n');
	    fw.write("NumDataPoints: " + constant.numDataPoints + '\r' + '\n');
	    fw.write("DimMap: ");
	    for (int i = 0; i <= current.enabledDim; i++) {
		fw.write("" + current.dimMap[i]);
		if (i != current.enabledDim)
		    fw.write(", ");
		else
		    fw.write("" + '\r' + '\n');
	    }
	    fw.write("ListMap: ");
	    for (int i = 0; i < constant.dataDim; i++) {
		fw.write("" + current.listMap[i]);
		if (i != constant.dataDim - 1)
		    fw.write(", ");
		else
		    fw.write("" + '\r' + '\n');
	    }
	    fw.write("DimLabels: ");
	    for (int i = 0; i < constant.dataDim; i++) {
		fw.write("\"" + constant.dimLabel[i] + "\"");
		if (i != constant.dataDim - 1)
		    fw.write(", ");
		else
		    fw.write("" + '\r' + '\n');
	    }
	    fw.write("CurrentLayout: \"" + 
		layoutPlugins[current.layoutType].getName() + '"' + '\r' + '\n');
	    int numPluginsWithOptions = 0;
	    for (int i = 0; i < numLayoutPlugins; i++) {
		if (layoutPlugins[i].getNumOptions() > 0)
		    numPluginsWithOptions++;
	    }
	    fw.write("LayoutPluginsData: " + numPluginsWithOptions + '\r' + '\n');
	    for (int i = 0; i < numLayoutPlugins; i++) {
		int numOptions = layoutPlugins[i].getNumOptions();
		if (numOptions > 0) {
		    fw.write("\"" + layoutPlugins[i].getName() + "\", " + numOptions + ", ");
		    for (int j = 0; j < numOptions; j++) {
			fw.write("\"" + layoutPlugins[i].getOptionName(j) + "\", ");
			if (layoutPlugins[i].getOptionValue(j))
			    fw.write("1");
			else
			    fw.write("0");
			if (j == numOptions - 1)
			    fw.write("" + '\r' + '\n');
			else
			    fw.write(", ");
		    }
		}
	    }
	    fw.write("CurrentGlyph: \"" + 
		glyphPlugins[current.glyphType].getName() + '"' + '\r' + '\n');
	    numPluginsWithOptions = 0;
	    for (int i = 0; i < numGlyphPlugins; i++) {
		if (glyphPlugins[i].getNumOptions() > 0)
		    numPluginsWithOptions++;
	    }
	    fw.write("GlyphPluginsData: " + numPluginsWithOptions + '\r' + '\n');
	    for (int i = 0; i < numGlyphPlugins; i++) {
		int numOptions = glyphPlugins[i].getNumOptions();
		if (numOptions > 0) {
		    fw.write("\"" + glyphPlugins[i].getName() + "\", " + numOptions + ", ");
		    for (int j = 0; j < numOptions; j++) {
			fw.write("\"" + glyphPlugins[i].getOptionName(j) + "\", ");
			if (glyphPlugins[i].getOptionValue(j))
			    fw.write("1");
			else
			    fw.write("0");
			if (j == numOptions - 1)
			    fw.write("" + '\r' + '\n');
			else
			    fw.write(", ");
		    }
		}
	    }
	    fw.write("Data: " + '\r' + '\n');
	    for (int i = 0; i < constant.numDataPoints; i++) {
		for (int j = 0; j <= constant.dataDim; j++) {
		    fw.write("" + constant.data[i][j]);
		    if (j != constant.dataDim)
			fw.write(", ");
		    else
			fw.write("" + '\r' + '\n');
		}
	    }

	    fw.close();
	} catch (Exception e) {
	    System.err.println("Couldn't save data:" + e);
	    return;
	}

	// Original state now equals current state
	original.copyFrom(current, constant.dataDim);
    } 

    // Default no data settings
    public ViewerState() {
	init();
    }

    void init() {
	current = new ViewerAdjustableState();
	original = new ViewerAdjustableState();
	constant = new ViewerStaticState();
	derived = new ViewerDerivedState();

	constant.tMin = 0.0;
	constant.tMax = 0.0;
	constant.tMinimumUnit = 1.0;

	current.cycleLength = 1.0;
	constant.tSpan = constant.tMax - constant.tMin + constant.tMinimumUnit;
	derived.numCycles = 1;

	constant.dataDim = 0;
	constant.numDataPoints = 0;
	constant.data = null;
	constant.normData = null;
	constant.dimLabel = null;

	current.enabledDim = constant.dataDim;
	current.dimMap = new int[constant.dataDim + 1];
	current.listMap = new int[constant.dataDim];

	current.tVal = constant.tMin;
	current.visibleCycles = 1;
	derived.tInterval = 1.0;
	derived.tVal2 = 0.0;
	current.glyphScale = 1.0;

	numLayoutPlugins = numGlyphPlugins = 0;
	layoutPlugins = null;
	glyphPlugins = null;
	current.glyphType = 0;
	current.layoutType = 0;
	layoutTypeDirty = true;

	current.cursorX = -1;
	current.cursorY = -1;
	
	current.brushOffsetX = -1;
	current.brushOffset = -1.0;

	current.autoNormalize = false;
	current.showLabels = false;
	current.showGuidelines = false;
	current.fillGlyphs = false;
	current.colorPerGlyph = false;

	fileName = null;
	original.copyFrom(current, constant.dataDim);
    }

    static double massageDouble(double val, double factor) {
	 return (Math.rint(val * factor) / factor);
    }

    public double getTVal() {
	return current.tVal;
    }

    public double getGlyphScale() {
	return current.glyphScale;
    }

    public double getCycleLength() {
	return current.cycleLength;
    }

    public double getBrushOffset() {
	return current.brushOffset;
    }

    public void setBrushOffset(double offset) {
	current.brushOffset = offset;
    }

    public int getBrushOffsetX() {
	return current.brushOffsetX;
    }

    public int getVisibleCycles() {
	return current.visibleCycles;
    }

    public int getEnabledDim() {
	return current.enabledDim;
    }

    public int getCursorX() {
	return current.cursorX;
    }

    public int getCursorY() {
	return current.cursorY;
    }

    public int getDimMap(int i) {
	return current.dimMap[i];
    }

    public boolean getFillGlyphs() {
	return current.fillGlyphs;
    }

    public boolean getColorPerGlyph() {
	return current.colorPerGlyph;
    }

    public boolean getShowLabels() {
	return current.showLabels;
    }

    public boolean getShowGuidelines() {
	return current.showGuidelines;
    }

    public boolean getAutoNormalize() {
	return current.autoNormalize;
    }

    public double getTVal2() {
	return derived.tVal2;
    }

    public double getTInterval() {
	return derived.tInterval;
    }

    public double getTMin() {
	return constant.tMin;
    }

    public int getNumDataPoints() {
	return constant.numDataPoints;
    }

    public double[] getNormData(int i) {
	return constant.normData[i];
    }

    public DataGlyph getGlyphPlugin() {
	return glyphPlugins[current.glyphType];
    }
}

class ViewerAdjustableState {
    double	cycleLength;
    int		visibleCycles;

    int		enabledDim;
    int		dimMap[];
    int		listMap[];

    double	tVal;

    double	glyphScale;
    int		glyphType;
    int		layoutType;

    int		cursorX;
    int		cursorY;

    int		brushOffsetX;
    double	brushOffset;

    boolean	autoNormalize;
    boolean	showLabels;
    boolean	showGuidelines;
    boolean	fillGlyphs;
    boolean	colorPerGlyph;

    void copyFrom(ViewerAdjustableState that, int dataDim) {
	cycleLength = that.cycleLength;
	visibleCycles = that.visibleCycles;

	enabledDim = that.enabledDim;

	listMap = new int[dataDim];
	for (int i = 0; i < dataDim; i++)
	    listMap[i] = that.listMap[i];

	dimMap = new int[dataDim + 1];
	for (int i = 0; i <= dataDim; i++) {
	    dimMap[i] = that.dimMap[i];
	}

	tVal = that.tVal;

	glyphScale = that.glyphScale;
	glyphType = that.glyphType;
	layoutType = that.layoutType;

	cursorX = -1;
	cursorY = -1;

	brushOffsetX = that.brushOffsetX;
	brushOffset = that.brushOffset;

	showLabels = that.showLabels;
	autoNormalize = that.autoNormalize;
	showGuidelines = that.showGuidelines;
	fillGlyphs = that.fillGlyphs;
	colorPerGlyph = that.colorPerGlyph;

	// TODO: Copy of plugin state!
    }

    boolean equalTo(ViewerAdjustableState that, int dataDim) {
	boolean breakOut = false;

	do {
	    if (cycleLength != that.cycleLength)
		break;

	    if (visibleCycles != that.visibleCycles)
		break;

	    if (enabledDim != that.enabledDim)
		break;

	    for (int i = 0; i <= enabledDim; i++) {
		if (dimMap[i] != that.dimMap[i]) {
		    breakOut = true;
		    break;
		}
	    }
	    if (breakOut)
		break;

	    for (int i = 0; i < dataDim; i++) {
		if (listMap[i] != that.listMap[i]) {
		    breakOut = true;
		    break;
		}
	    }
	    if (breakOut)
		break;

	    if (tVal != that.tVal)
		break;

	    if (glyphScale != that.glyphScale)
		break;

	    if (glyphType != that.glyphType)
		break;

	    if (layoutType != that.layoutType)
		break;

	    if (brushOffset != that.brushOffset)
		break;

	    if (autoNormalize != that.autoNormalize)
		break;

	    if (showLabels != that.showLabels)
		break;

	    if (showGuidelines != that.showGuidelines)
		break;

	    if (fillGlyphs != that.fillGlyphs)
		break;

	    if (colorPerGlyph != that.colorPerGlyph)
		break;

	    // TODO: Comparison of plugin state!

	    return true;
	} while (false);

	return false;
    }
}

class ViewerStaticState {
    double		tMin;
    double		tMax;
    double		tMinimumUnit;
    double		tSpan;

    int			numDataPoints;
    int			dataDim;
    double		data[][];
    double		normData[][];
    String		dimLabel[];
}

class ViewerDerivedState {
    double		tVal2;
    double		tInterval;
    int			numCycles;
}
