//
// BarGlyphPlugin.java
//
// This file provides the code used to extend the DataGlyph class to
// render a data point as a bar glyph.  Everything specific to bar
// glyphs should be contained within this file, including:
//
//  * How to draw the glyph within the view canvas,
//  * How to draw the glyph within the glyph legend canvas,
//  * How to calculate the offset for labels
//  * What glyph specific options exist
//

package spg.plugins;

import spg.*;
import java.awt.*;

public class BarGlyphPlugin extends DataGlyph {
    private static double MINIMUM = 0.01;		// Corresponds to minimum bar height

    public BarGlyphPlugin() {
	super("Bar Glyph");

	numOptions = 1;
	optionFlags = new boolean[numOptions];
	optionNames = new String[numOptions];
	optionFlags[0] = true;
	optionNames[0] = "Common Baseline";
    }

    public int getLabelOffset(double halfGlyphSize, double leastValue, double greatestValue,
	double dataPoint[], ViewerState global)
    {
	int lowestPoint = 0;
	double radius;
	if (optionFlags[0]) {
	    int tallestDim = (int)Math.rint(halfGlyphSize * 0.2);
	    lowestPoint = 0 - tallestDim;
	} else {
	    int thisY;
	    for (int j = 0; j < global.getEnabledDim(); j++) 
	    {
		radius = halfGlyphSize * ((0.5 - MINIMUM) *
		    (dataPoint[global.getDimMap(j + 1)] - leastValue) + MINIMUM);
		thisY = (int)Math.rint(radius * -1.0);
		if (thisY < lowestPoint)
		    lowestPoint = thisY;
	    }
	}

	return lowestPoint;
    }

    public void drawLegend(Graphics g, int sideLength, ViewerState global)
    {
	int labelX[] = new int[global.getEnabledDim()];
	int labelY[] = new int[global.getEnabledDim()];
	int barWidth, barHeight;
	int barX, barY;
	int labelWidth, labelHeight;
	boolean bigSmallToggle = true;

	labelHeight = legendFM.getAscent();
	barWidth = (sideLength - 10) / global.getEnabledDim();
	for (int j = 0; j < global.getEnabledDim(); j++) {
	    barX = 5 + (j * barWidth);
	    if (optionFlags[0]) {
		barY = 20;
		if (bigSmallToggle) {
		    barY = 20;
		    barHeight = sideLength - 40;
		    labelY[j] = barY - 5;
		} else {
		    barY = 40;
		    barHeight = sideLength - 60;
		    labelY[j] = sideLength - 5;
		}
	    } else {
		if (bigSmallToggle) {
		    barY = 20;
		    barHeight = sideLength - 40;
		    labelY[j] = barY - 5;
		} else {
		    barY = 30;
		    barHeight = sideLength - 60;
		    labelY[j] = sideLength - 5;
		}
	    }
	    bigSmallToggle = !bigSmallToggle;

	    labelWidth = legendFM.stringWidth("" + (j + 1));
	    labelX[j] = barX + ((barWidth - labelWidth) / 2);

	    if (!global.getColorPerGlyph())
		g.setColor(Color.getHSBColor((float)j / 
		    (float)global.getEnabledDim(), 
		    (float)1.0, (float)0.8));
	    else if (global.getFillGlyphs())
		g.setColor(Color.lightGray);
	    else
		g.setColor(Color.darkGray);
            
	    if (global.getFillGlyphs()) {
		g.fillRect(barX, barY, barWidth, barHeight);
		g.setColor(Color.darkGray);
		g.drawRect(barX, barY, barWidth, barHeight);
            } else
		g.drawRect(barX, barY, barWidth, barHeight);
	}

	// Draw dimension numbers
	g.setColor(Color.black);
	g.setFont(legendFont);
	for (int j = 0; j < global.getEnabledDim(); j++) {
	    g.drawString("" + (j + 1), labelX[j], labelY[j]);
	}
    }
    
    public void drawGlyph(Graphics g, int glyphX, int glyphY, double halfGlyphSize, 
	double leastValue, double greatestValue, double dataPoint[], ViewerState global)
    {
	Color glyphColor;
	int barX, barY, tallestDim = 0;
	int barWidth, halfBarHeight;

	glyphColor = g.getColor();

	barWidth = (int)Math.rint((halfGlyphSize * greatestValue) / 
            ((double)global.getEnabledDim() * global.getGlyphScale()));
	if (optionFlags[0])
	    tallestDim = (int)Math.rint(halfGlyphSize * 0.2);
	for (int j = 0; j < global.getEnabledDim(); j++) {
	    halfBarHeight = (int)Math.rint(halfGlyphSize * ((0.5 - MINIMUM) * 
		(dataPoint[global.getDimMap(j + 1)] - leastValue) + MINIMUM));
            barX = glyphX - (int)Math.rint((halfGlyphSize /
                global.getGlyphScale()) * greatestValue * 0.5) +
                (j * barWidth);
	    barY = optionFlags[0] ? (glyphY + tallestDim - (halfBarHeight * 2)) : 
		(glyphY - halfBarHeight);

	    if (!global.getColorPerGlyph())
		g.setColor(Color.getHSBColor((float)j /
		    (float)global.getEnabledDim(), 
                    (float)1.0, (float)0.8));

	    if (global.getFillGlyphs()) {
		if (global.getColorPerGlyph())
		    g.setColor(glyphColor);
                g.fillRect(barX, barY, barWidth, halfBarHeight * 2);
                g.setColor(Color.darkGray);
                g.drawRect(barX, barY, barWidth, halfBarHeight * 2);
	    } else {
		g.drawRect(barX, barY, barWidth, halfBarHeight * 2);
	    }
	}
    }
}
