//
// PieGlyphPlugin.java
//
// This file provides the code used to extend the DataGlyph class to
// render a data point as a pie glyph.  Everything specific to pie
// glyphs should be contained within this file, including:
//
//  * How to draw the glyph within the view canvas,
//  * How to draw the glyph within the glyph legend canvas,
//  * How to calculate the offset for labels
//  * What glyph specific options exist
//

package spg.plugins;

import spg.*;
import java.awt.*;

public class PieGlyphPlugin extends DataGlyph {
    private static double MINIMUM = 0.01;	    // Corresponds to minimum pie slice

    private double radius, angle, dA;
    private double invDimSum;
    private int lowestPoint, radiusInt;
    private int edgeX[], edgeY[];
    private static double EPSILON = 0.2;

    public PieGlyphPlugin() {
	super("Pie Glyph");

	radius = angle = dA = invDimSum = 0.0;
	lowestPoint = radiusInt = 0;
	edgeX = edgeY = null;
    }

    public int getLabelOffset(double halfGlyphSize, double leastValue, double greatestValue,
	double dataPoint[], ViewerState global)
    {
	lowestPoint = (int)Math.rint(-0.8 * halfGlyphSize * 
	    (greatestValue - leastValue));

	return lowestPoint;
    }

    public void drawLegend(Graphics g, int sideLength, ViewerState global)
    {
	double radius = (double)((sideLength / 2) - 20) - 0.5;
	double angle, dA;
	int dAInt;
	int labelX[] = new int[global.getEnabledDim()];
	int labelY[] = new int[global.getEnabledDim()];
	int edgeX[] = new int[global.getEnabledDim()];
	int edgeY[] = new int[global.getEnabledDim()];
	int labelWidth, labelHeight;
	int xOrigin, yOrigin;

	xOrigin = yOrigin = sideLength / 2;

	labelHeight = legendFM.getAscent();
	dA = (2.0 * Math.PI) / (double)global.getEnabledDim();
	for (int j = 0; j < global.getEnabledDim(); j++) {
	    angle = (Math.PI * 0.5) - (dA * (double)j);
	    edgeX[j] = xOrigin + (int)Math.rint(radius * 
		Math.cos(angle + (dA * 0.5)));
	    edgeY[j] = yOrigin - (int)Math.rint(radius * 
		Math.sin(angle + (dA * 0.5)));

	    labelWidth = legendFM.stringWidth("" + (j + 1));
	    labelX[j] = xOrigin + (int)Math.rint(((radius + 5.5) * Math.cos(angle)) +
		(((double)labelWidth / 2.0) * (Math.cos(angle) - 1.0)));
	    labelY[j] = yOrigin - (int)Math.ceil(((radius + 5.5) * Math.sin(angle)) +
		(((double)labelHeight / 2.0) * (Math.sin(angle) - 1.0)));
	}

	dA = 360.0 / (double)global.getEnabledDim();
	dAInt = 0 - (int)Math.rint(dA);
	for (int j = 0; j < global.getEnabledDim(); j++) {
	    if (!global.getColorPerGlyph())
		g.setColor(Color.getHSBColor((float)j /
		    (float)global.getEnabledDim(),
		    (float)1.0, (float)0.8));
	    else if (global.getFillGlyphs())
		g.setColor(Color.lightGray);
	    else
		g.setColor(Color.darkGray);

	    angle = 90.0 - (dA * ((double)j - 0.5));
	    if (global.getFillGlyphs()) {
		g.fillArc(20, 20, 
		    sideLength - 40, sideLength - 40,
		    (int)Math.rint(angle + EPSILON),
		    (int)Math.rint(0.0 - dA - EPSILON - EPSILON));
	    } else {
		g.drawArc(20, 20, 
		    sideLength - 40, sideLength - 40,
		    (int)Math.rint(angle), dAInt);
		if (global.getEnabledDim() > 1) {
		    if (!global.getColorPerGlyph())
			g.setColor(Color.lightGray);
		    g.drawLine(xOrigin, yOrigin,
			edgeX[j], edgeY[j]);
		}
	    }
	}

	if (global.getFillGlyphs()) {
	    g.setColor(Color.darkGray);

	    if (global.getEnabledDim() > 1)
		for (int j = 0; j < global.getEnabledDim(); j++)
		    g.drawLine(xOrigin, yOrigin, 
			edgeX[j], edgeY[j]);

	    g.drawOval(20, 20, sideLength - 40, sideLength - 40);
	}

	// Draw dimension numbers
	g.setColor(Color.black);
	g.setFont(legendFont);
	for (int j = 0; j < global.getEnabledDim(); j++)
	    g.drawString("" + (j + 1), labelX[j], labelY[j]);
    }
    
    public void drawGlyph(Graphics g, int glyphX, int glyphY, double halfGlyphSize, 
	double leastValue, double greatestValue, double dataPoint[], ViewerState global)
    {
	edgeX = new int[global.getEnabledDim()];
	edgeY = new int[global.getEnabledDim()];

	invDimSum = 0.0;
	for (int j = 0; j < global.getEnabledDim(); j++)
	    invDimSum += dataPoint[global.getDimMap(j + 1)];
	invDimSum = (2.0 * Math.PI) / invDimSum;

	if (global.getAutoNormalize())
	    radius = halfGlyphSize * (greatestValue - leastValue) * 0.8;
	else
	    radius = halfGlyphSize * 0.8;
	radiusInt = (int)Math.rint(radius);
	angle = (Math.PI * 0.5) + 
	    (dataPoint[global.getDimMap(1)] * invDimSum * 0.5);
	for (int j = 0; j < global.getEnabledDim(); j++) {
	    angle -= (dataPoint[global.getDimMap(j + 1)] * invDimSum);
	    edgeX[j] = glyphX + (int)Math.rint(radius * Math.cos(angle));
	    edgeY[j] = glyphY - (int)Math.rint(radius * Math.sin(angle));
        }
	invDimSum = (invDimSum * 180.0) / Math.PI;
	angle = 90.0 + 
	    (dataPoint[global.getDimMap(1)] * invDimSum * 0.5);
	for (int j = 0; j < global.getEnabledDim(); j++) {
	    dA = dataPoint[global.getDimMap(j + 1)] * invDimSum;
	    if (!global.getColorPerGlyph())
		g.setColor(Color.getHSBColor((float)j /
		    (float)global.getEnabledDim(),
		    (float)1.0, (float)0.8));

	    if (global.getFillGlyphs())
	    {
		g.fillArc(glyphX - radiusInt, 
		    glyphY - radiusInt, 
		    radiusInt * 2, radiusInt * 2,
		    (int)Math.rint(angle + EPSILON), (int)Math.rint(0.0 - dA - EPSILON - EPSILON));
            } else {
		g.drawArc(glyphX - radiusInt, 
		    glyphY - radiusInt, 
		    radiusInt * 2, radiusInt * 2,
		    (int)Math.rint(angle), (int)Math.rint(0.0 - dA));
            }

	    angle -= dA;
	}

	if (global.getFillGlyphs()) {
	    g.setColor(Color.darkGray);
	    g.drawOval(glyphX - radiusInt, 
		glyphY - radiusInt, 
		radiusInt * 2, radiusInt * 2);
	} else if (!global.getColorPerGlyph())
	    g.setColor(Color.lightGray);

	if (global.getEnabledDim() > 1)
	    for (int j = 0; j < global.getEnabledDim(); j++)
		g.drawLine(glyphX, glyphY,
		    edgeX[j], edgeY[j]);
    }
}
