//
// SpiralLayoutPlugin.java
//
// This file provides the code used to extend the DataLayout class to
// map the data onto a spiral.
//

package spg.plugins;

import spg.*;
import java.awt.*;

public class SpiralLayoutPlugin extends DataLayout {
    private static int xAH[] = {0, 5, 10};
    private static int yAH[] = {0, -8, 0};
    private int xOrigin, yOrigin, glyphX, glyphY;
    private int halfSideLength;
    private int ringSpacing, rS2, rSDiv4, rSDiv2, rS3Div4, rS3Div2;
    private int iRS, iRS2, dX, dY;
    private int halfGlyphSizeInt;
    private double radius, angle, dR;
    private double halfGlyphSize, cycleNumber, cycleOffset;
    private double tVal, dataPoint[];
    private Polygon arrowHead;

    public SpiralLayoutPlugin() {
	super("Spiral Layout");

	numOptions = 2;
	optionFlags = new boolean[numOptions];
	optionNames = new String[numOptions];
	optionFlags[0] = true;
	optionNames[0] = "Show Spiral";
	optionFlags[1] = true;
	optionNames[1] = "Spiral Outward";

	xOrigin = yOrigin = glyphX = glyphY = halfSideLength = 0;
	ringSpacing = rS2 = rSDiv4 = rSDiv2 = rS3Div4 = rS3Div2 = 0;
	iRS = iRS2 = dX = dY = halfGlyphSizeInt = 0;
	radius = angle = dR = tVal = 0.0;
	cycleNumber = cycleOffset = halfGlyphSize = 0.0;
	arrowHead = null;
	dataPoint = null;
    }

    public Rectangle getLayoutBounds(Dimension d) {
	if (d.height > d.width) {
	   bounds.width = bounds.height = d.width - 20;
	   bounds.y = (d.height - bounds.height) / 2;
	   bounds.x = 10;
	} else {
	   bounds.width = bounds.height = d.height - 20;
	   bounds.y = 10;
	   bounds.x = (d.width - bounds.width) / 2;
	}

	return bounds;
    }

    public void drawLayout(Graphics g, ViewerState global) {
	currentGlyphType = global.getGlyphPlugin();

	halfSideLength = bounds.width / 2;
	xOrigin = bounds.x + halfSideLength;
	yOrigin = bounds.y + halfSideLength;

	// Draw guidelines
	if (global.getShowGuidelines()) {
	    g.setColor(Color.lightGray);

	    g.drawLine(bounds.x + 1, yOrigin, 
		bounds.x + bounds.width - 1, yOrigin);
	    g.drawLine(xOrigin, bounds.y + 1, 
		xOrigin, bounds.y + bounds.width - 1);
	    g.drawLine(bounds.x + 1, bounds.y + 1, 
		bounds.x + bounds.width - 1, 
		bounds.y + bounds.width - 1);
	    g.drawLine(bounds.x + bounds.width - 1, 
		bounds.y + 1, bounds.x + 1,
		bounds.y + bounds.width - 1);
	}

	ringSpacing = bounds.width / (2 * global.getVisibleCycles() + 3);
	rS2 = ringSpacing * 2;
	rSDiv4 = ringSpacing / 4;
	rSDiv2 = rSDiv4 * 2;
	rS3Div4 = rSDiv4 * 3;
	rS3Div2 = rSDiv2 * 3;

	// Draw spiral
	if (optionFlags[0]) {
	    g.setColor(Color.pink);
	    for (int i = 1; i <= global.getVisibleCycles(); i++) {
		iRS = i * ringSpacing;
		iRS2 = i * rS2;

		g.drawArc(xOrigin - iRS - rSDiv4, 
		    yOrigin - iRS,
		    iRS2 + rSDiv2, 
		    iRS2, 90, -90);
		g.drawArc(xOrigin - iRS - rSDiv4, 
		    yOrigin - iRS - rSDiv2, 
		    iRS2 + rSDiv2,
		    iRS2 + ringSpacing, 0, -90);
		g.drawArc(xOrigin - iRS - rS3Div4, 
		    yOrigin - iRS - rSDiv2, 
		    iRS2 + rS3Div2, 
		    iRS2 + ringSpacing, -90, -90);
		g.drawArc(xOrigin - iRS - rS3Div4,
		    yOrigin - iRS - ringSpacing, 
		    iRS2 + rS3Div2,
		    iRS2 + rS2, -180, -90);
	    }

	    // Draw directional indicators
	    iRS = (global.getVisibleCycles() + 1) * ringSpacing; 
	    iRS2 = iRS * 2;
	    g.drawArc(xOrigin - iRS, 
		yOrigin - iRS, 
		iRS2, 
		iRS2, 90, -30);
	    g.drawArc(xOrigin - ringSpacing,
		yOrigin - ringSpacing, 
		rS2,
		rS2, -240, -30);
	
	    arrowHead = new Polygon(xAH, yAH, 3);
	    if (optionFlags[1]) {
		// We're drawing the arrow head on the outermost ring

		dX = xOrigin + (int)Math.rint((double)(iRS - 5) * 0.5); 
		dY = yOrigin - (int)Math.rint((double)(iRS - 5) * 0.866);
	    } else {
		// We're drawing the arrow head on the innermost ring

		dX = xOrigin - (int)Math.rint((double)(ringSpacing - 5) * 0.5) - 10;
		dY = yOrigin - (int)Math.rint((double)(ringSpacing - 5) * 0.866);
	    }
	    arrowHead.translate(dX, dY);
	    g.fillPolygon(arrowHead);
	}

	// Determine size of glyphs
	halfGlyphSize = (global.getGlyphScale() * (double)ringSpacing * 0.5) / (greatestValue - leastValue);
	halfGlyphSizeInt = (int)Math.rint(halfGlyphSize);

	// Draw the data
	for (int i = 0; i < global.getNumDataPoints(); i++) {
	    dataPoint = global.getNormData(i);
	    tVal = dataPoint[global.getDimMap(0)];

	    // Ignore data that is outside of visible interval
	    if ((tVal < global.getTVal()) ||
		(tVal > global.getTVal2())) {
		  continue;
	    }

	    // Determine glyph origin (position for center of glyph)
	    cycleOffset = Math.IEEEremainder((tVal - global.getTVal()), 
		global.getCycleLength());
	    if (cycleOffset < 0.0)
		  cycleOffset += global.getCycleLength();
	    cycleNumber = Math.floor((tVal - global.getTVal()) / 
		global.getCycleLength()) + 1.0;
	    dR = cycleOffset / global.getCycleLength();

	    if (global.getColorPerGlyph()) {
		// Determine glyph color based on cycle offset in data space
		cycleOffset = Math.IEEEremainder((tVal - global.getTMin()), 
		    global.getCycleLength());
		if (cycleOffset < 0.0)
		    cycleOffset += global.getCycleLength();
		g.setColor(Color.getHSBColor(
		    (float)(cycleOffset / global.getCycleLength()), 
                    (float)1.0, (float)0.8));
	    }

	    if (optionFlags[1]) {
		angle = (Math.PI * 0.5) - (2.0 * Math.PI * dR);
		radius = (cycleNumber + dR) * (double)ringSpacing;
	    } else {
		angle = (Math.PI * 0.5) + (2.0 * Math.PI * dR);
		radius = ((double)global.getVisibleCycles() + 2.0 - cycleNumber - dR) * (double)ringSpacing;
	    }

	    glyphX = xOrigin + (int)Math.rint(radius * Math.cos(angle));
	    glyphY = yOrigin - (int)Math.rint(radius * Math.sin(angle));

	    // Draw the apropriate glyph
	    currentGlyphType.drawGlyph(g, glyphX, glyphY, halfGlyphSize, 
		leastValue, greatestValue, dataPoint, global);
	}

	// Draw data labels if necessary
	if (global.getShowLabels()) {
	    g.setFont(labelFont);
	    g.setColor(Color.black);
	    int labelHeight = labelFM.getAscent();
	    int lowestPoint, thisY;
	    int startCheck;

	    for (int i = 0; i < global.getNumDataPoints(); i++) {
		dataPoint = global.getNormData(i);
		tVal = dataPoint[global.getDimMap(0)];

		// Ignore data that is outside of visible interval
		if ((tVal < global.getTVal()) ||
		    (tVal > global.getTVal2()))
		    continue;

		// Determine glyph origin (position for center of glyph)
		cycleOffset = Math.IEEEremainder((tVal - global.getTVal()), 
		    global.getCycleLength());
		if (cycleOffset < 0.0)
		    cycleOffset += global.getCycleLength();
		cycleNumber = Math.floor((tVal - global.getTVal()) / 
		    global.getCycleLength()) + 1.0;
		dR = cycleOffset / global.getCycleLength();

		if (optionFlags[1]) {
		    angle = (Math.PI * 0.5) - (2.0 * Math.PI * dR);
		    radius = (cycleNumber + dR) * (double)ringSpacing;
		} else {
		    angle = (Math.PI * 0.5) + (2.0 * Math.PI * dR);
		    radius = ((double)global.getVisibleCycles() + 2.0 - cycleNumber - dR) * (double)ringSpacing;
		}

		glyphX = xOrigin + (int)Math.rint(radius * Math.cos(angle));
		glyphY = yOrigin - (int)Math.rint(radius * Math.sin(angle));

		// Get lower radius of data point
		lowestPoint = currentGlyphType.getLabelOffset(halfGlyphSize, leastValue, 
		    greatestValue, dataPoint, global);

		// Draw the label
		g.drawString("" + tVal, 
		    glyphX - (labelFM.stringWidth("" + tVal) / 2),
		    glyphY - lowestPoint + labelHeight);
	    }
	}

	// If the cursor is hovering over a glyph,
	// dump its raw values...
	if (global.getCursorX() != -1) {
	    for (int i = 0; i < global.getNumDataPoints(); i++) {
		dataPoint = global.getNormData(i);
		tVal = dataPoint[global.getDimMap(0)];

		// Ignore data that is outside of visible interval
		if ((tVal < global.getTVal()) ||
		    (tVal > global.getTVal2()))
		    continue;

		// Determine glyph origin (position for center of glyph)
		cycleOffset = Math.IEEEremainder((tVal - global.getTVal()), 
		    global.getCycleLength());
		if (cycleOffset < 0.0)
		    cycleOffset += global.getCycleLength();
		cycleNumber = Math.floor((tVal - global.getTVal()) / 
		    global.getCycleLength()) + 1.0;
		dR = cycleOffset / global.getCycleLength();

		if (optionFlags[1]) {
		    angle = (Math.PI * 0.5) - (2.0 * Math.PI * dR);
		    radius = (cycleNumber + dR) * (double)ringSpacing;
		} else {
		    angle = (Math.PI * 0.5) + (2.0 * Math.PI * dR);
		    radius = ((double)global.getVisibleCycles() + 2.0 - cycleNumber - dR) * (double)ringSpacing;
		}

		glyphX = xOrigin + (int)Math.rint(radius * Math.cos(angle));
		glyphY = yOrigin - (int)Math.rint(radius * Math.sin(angle));

		if ((global.getCursorX() < (glyphX + (halfGlyphSizeInt / 2))) &&
		    (global.getCursorX() > (glyphX - (halfGlyphSizeInt / 2))) &&
		    (global.getCursorY() < (glyphY + (halfGlyphSizeInt / 2))) &&
		    (global.getCursorY() > (glyphY - (halfGlyphSizeInt / 2))))
		{
		    int dumpAtX = global.getCursorX() + (halfGlyphSizeInt / 6) + 10;
		    int dumpAtY = global.getCursorY() + (halfGlyphSizeInt / 6) + 10;

		    drawDataDump(g, dumpAtX, dumpAtY, bounds.x + bounds.width, 
			bounds.y + bounds.height, i, global);
		    break;
		}
	    }
	}
    }
}