//
// StarGlyphPlugin.java
//
// This file provides the code used to extend the DataGlyph class to
// render a data point as a star glyph.  Everything specific to star
// glyphs should be contained within this file, including:
//
//  * How to draw the glyph within the view canvas,
//  * How to draw the glyph within the glyph legend canvas,
//  * How to calculate the offset for labels
//  * What glyph specific options exist
//

package spg.plugins;

import spg.*;
import java.awt.*;

public class StarGlyphPlugin extends DataGlyph {
    private static double MINIMUM = 0.03;	// Corresponds to minimum ray length

    private double dA, angle, radius;
    private double weight, lastRadius, firstRadius;
    private int thisY, lowestPoint, startCheck;
    private int edgeX[], edgeY[], midX[], midY[];
    private int quadX[];
    private int quadY[];

    public StarGlyphPlugin() {
	super("Star Glyph");

	dA = angle = radius = lastRadius = firstRadius = weight = 0.0;
	thisY = lowestPoint = startCheck;
	quadX = new int[4];
	quadY = new int[4];
	edgeX = edgeY = midX = midY = null;
    }

    public int getLabelOffset(double halfGlyphSize, double leastValue, double greatestValue,
	double dataPoint[], ViewerState global)
    {
	startCheck = (int)Math.rint(global.getEnabledDim() / 3.0);
	lowestPoint = 0;
	for (int j = startCheck; 
	    j <= global.getEnabledDim() - startCheck; j++) 
	{
	    angle = (Math.PI * 0.5) - (dA * (double)j);
	    radius = halfGlyphSize * ((1.0 - MINIMUM) *
		(dataPoint[global.getDimMap(j + 1)] - leastValue) + MINIMUM);
	    thisY = (int)Math.rint(radius * Math.sin(angle));
	    if (thisY < lowestPoint)
		lowestPoint = thisY;
	}

	return lowestPoint;
    }

    public void drawLegend(Graphics g, int sideLength, ViewerState global)
    {
	double radius = (double)((sideLength / 2) - 20);
	double angle, dA;
	int labelX[] = new int[global.getEnabledDim()];
	int labelY[] = new int[global.getEnabledDim()];
	int edgeX[] = new int[global.getEnabledDim()];
	int edgeY[] = new int[global.getEnabledDim()];
        int midX[] = new int[global.getEnabledDim()];
        int midY[] = new int[global.getEnabledDim()];
        int quadX[] = new int[4];
        int quadY[] = new int[4];
	int labelWidth, labelHeight;
	int xOrigin, yOrigin;

	xOrigin = yOrigin = sideLength / 2;

	labelHeight = legendFM.getAscent();
	dA = (2.0 * Math.PI) / (double)global.getEnabledDim();
	for (int j = 0; j < global.getEnabledDim(); j++) {
	    angle = (Math.PI * 0.5) - (dA * (double)j);
	    edgeX[j] = xOrigin + (int)Math.rint(radius * Math.cos(angle));
	    edgeY[j] = yOrigin - (int)Math.rint(radius * Math.sin(angle));
	    if (j > 0) {
		midX[j] = (edgeX[j - 1] + edgeX[j]) / 2;
		midY[j] = (edgeY[j - 1] + edgeY[j]) / 2;
            }

	    labelWidth = legendFM.stringWidth("" + (j + 1));
	    labelX[j] = xOrigin + (int)Math.rint(((radius + 5) * Math.cos(angle)) +
		(((double)labelWidth / 2.0) * (Math.cos(angle) - 1.0)));
	    labelY[j] = yOrigin - (int)Math.ceil(((radius + 5) * Math.sin(angle)) +
		(((double)labelHeight / 2.0) * (Math.sin(angle) - 1.0)));
	}
	midX[0] = (edgeX[global.getEnabledDim() - 1] + edgeX[0]) / 2;
	midY[0] = (edgeY[global.getEnabledDim() - 1] + edgeY[0]) / 2;


	for (int j = 0; j < global.getEnabledDim(); j++) {
	    if (!global.getColorPerGlyph())
		g.setColor(Color.getHSBColor((float)j /
		    (float)global.getEnabledDim(),
		    (float)1.0, (float)0.8));
	    else if (global.getFillGlyphs() &&
		global.getEnabledDim() > 2)
	    {
		g.setColor(Color.lightGray);
	    } else
		g.setColor(Color.darkGray);

	    if (global.getFillGlyphs() &&
		global.getEnabledDim() > 2) 
	    {
		quadX[0] = xOrigin; quadY[0] = yOrigin;
		quadX[2] = edgeX[j]; quadY[2] = edgeY[j];
		quadX[1] = midX[j]; quadY[1] = midY[j];
		quadX[3] = midX[(j + 1) % global.getEnabledDim()];
		quadY[3] = midY[(j + 1) % global.getEnabledDim()];
		g.fillPolygon(quadX, quadY, 4);
	    } else {
		g.drawLine(xOrigin, yOrigin, edgeX[j], edgeY[j]);
	    }
	}

	if (global.getFillGlyphs()) {
	    g.setColor(Color.darkGray);

	    if (global.getEnabledDim() > 2) {
		for (int j = 0; j < global.getEnabledDim(); j++)
		    g.drawLine(xOrigin, yOrigin, midX[j], midY[j]);
	    }
	}
	if (global.getEnabledDim() > 2) {
	    if (!global.getFillGlyphs())
		g.setColor(Color.lightGray);
	    g.drawPolygon(edgeX, edgeY, global.getEnabledDim());
	} else if (global.getEnabledDim() == 2) {
	    // Draw marker so we can distinguish between 2 dimensions
	    g.setColor(Color.darkGray);
	    g.drawLine(xOrigin - 2, yOrigin, xOrigin + 2, yOrigin);
	}

	// Draw dimension numbers
	g.setColor(Color.black);
	g.setFont(legendFont);
	for (int j = 0; j < global.getEnabledDim(); j++) {
	    g.drawString("" + (j + 1), labelX[j], labelY[j]);
	}
    }
    
    public void drawGlyph(Graphics g, int glyphX, int glyphY, double halfGlyphSize, 
	double leastValue, double greatestValue, double dataPoint[], ViewerState global) 
    {
        lastRadius = firstRadius = 0.0;
	edgeX = new int[global.getEnabledDim()];
	edgeY = new int[global.getEnabledDim()];
	midX = new int[global.getEnabledDim()];
	midY = new int[global.getEnabledDim()];

        dA = (2.0 * Math.PI) / (double)global.getEnabledDim();
	for (int j = 0; j < global.getEnabledDim(); j++) {
	    angle = (Math.PI * 0.5) - (dA * (double)j);
	    radius = halfGlyphSize * ((1.0 - MINIMUM) * (dataPoint[global.getDimMap(j + 1)] 
		- leastValue) + MINIMUM);
            edgeX[j] = glyphX + (int)Math.rint(radius * Math.cos(angle));
            edgeY[j] = glyphY - (int)Math.rint(radius * Math.sin(angle));
            if (j > 0) {
		if (lastRadius == 0.0 && radius == 0.0)
		    weight = 0.5;
		else {
		    weight = lastRadius * lastRadius;
		    weight /= (weight + (radius * radius));
		}
                midX[j] = (int)Math.rint(((double)edgeX[j-1] * (1.0 - weight)) +
                    ((double)edgeX[j] * weight));
                midY[j] = (int)Math.rint(((double)edgeY[j-1] * (1.0 - weight)) +
                    ((double)edgeY[j] * weight));
            } else {
                firstRadius = radius;
            }
            lastRadius = radius;
        }
	if (lastRadius == 0.0 && firstRadius == 0.0)
	    weight = 0.5;
	else {
	    weight = lastRadius * lastRadius;
	    weight /= (weight + (firstRadius * firstRadius));
	}
        midX[0] = (int)Math.rint(((double)edgeX[global.getEnabledDim() - 1] * (1.0 - weight)) +
            ((double)edgeX[0] * weight));
        midY[0] = (int)Math.rint(((double)edgeY[global.getEnabledDim() - 1] * (1.0 - weight)) +
            ((double)edgeY[0] * weight));

	for (int j = 0; j < global.getEnabledDim(); j++) {
	    if (!global.getColorPerGlyph())
		g.setColor(Color.getHSBColor((float)j / 
		    (float)global.getEnabledDim(), 
                    (float)1.0, (float)0.8));

	    if (global.getFillGlyphs() &&
		global.getEnabledDim() > 2)
	    {
		quadX[0] = glyphX; quadY[0] = glyphY;
		quadX[2] = edgeX[j]; quadY[2] = edgeY[j];
		quadX[1] = midX[j]; quadY[1] = midY[j];
		quadX[3] = midX[(j + 1) % global.getEnabledDim()];
		quadY[3] = midY[(j + 1) % global.getEnabledDim()];

		g.fillPolygon(quadX, quadY, 4);
            } else {
                g.drawLine(glyphX, glyphY, edgeX[j], edgeY[j]);
	    }
	}

	if (global.getFillGlyphs()) {
            g.setColor(Color.darkGray);

            if (global.getEnabledDim() > 2) {
                for (int j = 0; j < global.getEnabledDim(); j++)
                    g.drawLine(glyphX, glyphY, midX[j], midY[j]);
            }
	} else if (!global.getColorPerGlyph())
	    g.setColor(Color.lightGray);

	if (global.getEnabledDim() > 2)
            g.drawPolygon(edgeX, edgeY, global.getEnabledDim());
	else if (global.getEnabledDim() == 2)
	    // Draw marker so we can distinguish between 2 dimensions
	    g.drawLine(glyphX - 2, glyphY, glyphX + 2, glyphY);
    }
}

